/**
 * @copyright Cube Dev, Inc.
 * @license Apache-2.0
 * @fileoverview The `PostgresDriver` and related types declaration.
 */
import { Pool, PoolConfig, PoolClient, FieldDef } from 'pg';
import { TypeId, TypeFormat } from 'pg-types';
import { BaseDriver, DownloadQueryResultsOptions, DownloadTableMemoryData, DriverInterface, GenericDataBaseType, IndexesSQL, TableStructure, StreamOptions, StreamTableDataWithTypes, QueryOptions, DownloadQueryResultsResult } from '@cubejs-backend/base-driver';
import { QueryStream } from './QueryStream';
export type PostgresDriverConfiguration = Partial<PoolConfig> & {
    storeTimezone?: string;
    executionTimeout?: number;
    readOnly?: boolean;
    /**
     * The export bucket CSV file escape symbol.
     */
    exportBucketCsvEscapeSymbol?: string;
};
/**
 * Postgres driver class.
 */
export declare class PostgresDriver<Config extends PostgresDriverConfiguration = PostgresDriverConfiguration> extends BaseDriver implements DriverInterface {
    /**
     * Returns default concurrency value.
     */
    static getDefaultConcurrency(): number;
    private enabled;
    protected readonly pool: Pool;
    protected readonly config: Partial<Config>;
    /**
     * Class constructor.
     */
    constructor(config?: Partial<Config> & {
        /**
         * Data source name.
         */
        dataSource?: string;
        /**
         * Max pool size value for the [cube]<-->[db] pool.
         */
        maxPoolSize?: number;
        /**
         * Time to wait for a response from a connection after validation
         * request before determining it as not valid. Default - 10000 ms.
         */
        testConnectionTimeout?: number;
    });
    /**
     * The easiest way how to add additional configuration from env variables, because
     * you cannot call method in RedshiftDriver.constructor before super.
     */
    protected getInitialConfiguration(dataSource: string): Partial<PostgresDriverConfiguration>;
    protected getTypeParser: (dataTypeID: TypeId, format: TypeFormat | undefined) => ((val: string) => string) | ((val: any) => any);
    /**
     * It's not possible to detect user defined types via constant oids
     * For example HLL extensions is using CREATE TYPE HLL which will generate a new pg_type with different oids
     */
    protected userDefinedTypes: Record<string, string> | null;
    protected getPostgresTypeForField(dataTypeID: number): string | null;
    testConnection(): Promise<void>;
    protected loadUserDefinedTypes(conn: PoolClient): Promise<void>;
    protected prepareConnection(conn: PoolClient, options?: {
        executionTimeout: number;
    }): Promise<void>;
    protected mapFields(fields: FieldDef[]): {
        name: string;
        type: string;
    }[];
    streamQuery(sql: string, values: string[]): Promise<QueryStream>;
    stream(query: string, values: unknown[], { highWaterMark }: StreamOptions): Promise<StreamTableDataWithTypes>;
    protected queryResponse(query: string, values: unknown[]): Promise<import("pg").QueryResult<any>>;
    query<R = unknown>(query: string, values: unknown[], options?: QueryOptions): Promise<R[]>;
    downloadQueryResults(query: string, values: unknown[], options: DownloadQueryResultsOptions): Promise<DownloadQueryResultsResult>;
    toGenericType(columnType: string): GenericDataBaseType;
    readOnly(): boolean;
    uploadTableWithIndexes(table: string, columns: TableStructure, tableData: DownloadTableMemoryData, indexesSql: IndexesSQL): Promise<void>;
    release(): Promise<void>;
    param(paramIndex: number): string;
    fromGenericType(columnType: string): string;
}
//# sourceMappingURL=PostgresDriver.d.ts.map